/*________________________________________________________________________________________________

	Title: Convert to RNL Chromaticity
	'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
	Author: Jolyon Troscianko
	Date: 16/11/2018
............................................................................................................

This script converts from cone-catch quanta to receptor-noise-limited chromaticity space (as described
in Renoult et al 2017; Colour spaces in ecology and evolutionary biology, equations A1.23 to A 1.29.)

Euclidean distances measured in this colour space are identical to JND calculations.

This script uses the log-normal version of the RNL model, up to tetrachromacy.

--------------------------------------------------------------------------------------------------------------------------------
Requirements:

Input images should be cone-catch images with slices labelled according to the receptor channel.
The user is asked to input the Weber fraction specific to each channel.

Any channels with Weber fractions of zero are ignored
_________________________________________________________________________________________________
*/



import ij.*;
import ij.plugin.filter.PlugInFilter;
import ij.process.*;
import ij.gui.*;
import ij.plugin.*;
import java.awt.*;



public class Convert_to_RNL_Chromaticity implements PlugInFilter {

ImageStack stack;
	public int setup(String arg, ImagePlus imp) { 
	stack = imp.getStack(); 
	return DOES_32 + STACK_REQUIRED; 
	}



public void run(ImageProcessor ip) {


IJ.showStatus("Convert to RNL");


int w = stack.getWidth();
int h = stack.getHeight();
int nPoints = w*h;
int nDims = stack.getSize();

double[] weber = new double[nDims]; //

//for(int i = 0; i<nDims; i++)
//	weber[i] = 0.05;

String[] sliceNames = new String[nDims];
sliceNames = stack.getSliceLabels();
int nChannels = 0;
int[] chL = new int[nDims]; // channel lookup


	GenericDialog gd = new GenericDialog("Convert to RNL Chromaticity");

		for(int i=0; i<nDims; i++)
			gd.addNumericField("Weber_" + (i+1) + " " + sliceNames[i], weber[i], 3);	

	gd.showDialog();
	if (gd.wasCanceled())
		return;


for(int i=0; i<nDims; i++){
	weber[i] = gd.getNextNumber();
	if(weber[i] != 0.0){
		IJ.log("" + sliceNames[i] + " " + weber[i]);
		chL[nChannels] = i;
		nChannels++;
	}
}

if(nChannels > 4 || nChannels < 2){
	IJ.log("This tool only supports di, tri, or tetrachromatic images");
	return;
}

float[] pxs = new float[nPoints*nDims];
float[] pixels = new float[nPoints*nDims];

for(int i=0; i<nDims; i++){ // convert float pixel values to double-precision array
	pixels = (float[]) stack.getPixels(i+1);
	for(int k=0; k<nPoints; k++)
		pxs[k+nPoints*i] = pixels[k];
}


ImageStack outStack = new ImageStack(w, h);

/*
IJ.log("__________________________");

IJ.log("Weber 1: " + weber[chL[0]]);
IJ.log("Weber 2: " + weber[chL[1]]);
IJ.log("Weber 3: " + weber[chL[2]]);
*/


double c0 =0.0;
double c1 = 0.0;
double c2 = 0.0;
double c3 = 0.0;
double A = 0.0;
double a = 0.0;
double b = 0.0;
double c = 0.0;


float[] chrom = new float[nPoints];

//-----------X--------------
float[] x = new float[nPoints];
for(int i=0; i<nPoints; i++){
	c0 = (double) (pxs[i+nPoints*chL[0]]);
	c1 = (double) (pxs[i+nPoints*chL[1]]);
	x[i] = (float) (   Math.pow(1/(Math.pow(weber[chL[1]],2)+Math.pow(weber[chL[0]], 2)), 0.5)*(Math.log(c0)-Math.log(c1))       );		     
	chrom[i] += x[i] * x[i];
}

outStack.addSlice("X ("+sliceNames[0]+":"+sliceNames[1]+")", x);


//-----------Y--------------
if(nChannels > 2){
float[] y = new float[nPoints];
for(int i=0; i<nPoints; i++){
	c0 = (double) (pxs[i+nPoints*chL[0]]);
	c1 = (double) (pxs[i+nPoints*chL[1]]);
	c2 = (double) (pxs[i+nPoints*chL[2]]);
	y[i] = (float) (        Math.pow((Math.pow(weber[chL[1]] , 2)+Math.pow(weber[chL[0]] , 2))/(Math.pow(weber[chL[2]]*weber[chL[1]], 2)+Math.pow(weber[chL[2]]*weber[chL[0]], 2)+Math.pow(weber[chL[1]]*weber[chL[0]],2)) , 0.5)*(Math.log(c2)-(Math.log(c0)*(Math.pow(weber[chL[1]], 2)/(Math.pow(weber[chL[1]], 2)+Math.pow(weber[chL[0]],2)))+Math.log(c1)*(Math.pow(weber[chL[0]],2)/(Math.pow(weber[chL[1]], 2)+Math.pow(weber[chL[0]], 2)))))          );
	chrom[i] += y[i] * y[i];
}

outStack.addSlice("Y ("+sliceNames[0]+"+"+sliceNames[1]+":"+sliceNames[2]+")", y);
}

//-----------Z--------------
if(nChannels > 3){
float[] z = new float[nPoints];
for(int i=0; i<nPoints; i++){
	c0 = (double) (pxs[i+nPoints*chL[0]]);
	c1 = (double) (pxs[i+nPoints*chL[1]]);
	c2 = (double) (pxs[i+nPoints*chL[2]]);
	c3 = (double) (pxs[i+nPoints*chL[3]]);
	A = Math.pow((Math.pow(weber[chL[1]]*weber[chL[0]], 2)+Math.pow(weber[chL[2]]*weber[chL[0]], 2)+Math.pow(weber[chL[2]]*weber[chL[1]], 2))/(Math.pow(weber[chL[2]]*weber[chL[1]]*weber[chL[0]], 2)+Math.pow(weber[chL[3]]*weber[chL[1]]*weber[chL[0]], 2)+Math.pow(weber[chL[3]]*weber[chL[2]]*weber[chL[0]], 2)+Math.pow(weber[chL[3]]*weber[chL[2]]*weber[chL[1]], 2)),0.5);
	a = Math.pow(weber[chL[2]]*weber[chL[1]], 2)/(Math.pow(weber[chL[1]]*weber[chL[0]], 2)+Math.pow(weber[chL[2]]*weber[chL[1]], 2)+Math.pow(weber[chL[2]]*weber[chL[0]], 2));
	b = Math.pow(weber[chL[2]]*weber[chL[0]], 2)/(Math.pow(weber[chL[1]]*weber[chL[0]], 2)+Math.pow(weber[chL[2]]*weber[chL[1]], 2)+Math.pow(weber[chL[2]]*weber[chL[0]], 2));
	c = Math.pow(weber[chL[1]]*weber[chL[0]], 2)/(Math.pow(weber[chL[1]]*weber[chL[0]], 2)+Math.pow(weber[chL[2]]*weber[chL[1]], 2)+Math.pow(weber[chL[2]]*weber[chL[0]], 2));
	z[i] = (float) (       A*(Math.log(c3)-(a*Math.log(c0)+b*Math.log(c1)+c*Math.log(c2)))     );
	chrom[i] += z[i] * z[i];
}

outStack.addSlice("Z ("+sliceNames[0]+"+"+sliceNames[1]+"+"+sliceNames[2]+":"+sliceNames[3]+")", z);
}

for(int i=0; i<nPoints; i++)
	chrom[i] = (float) Math.pow(chrom[i], 0.5);

outStack.addSlice("Saturation", chrom);

new ImagePlus("RNL Chromaticity", outStack).show();




}
}
