/*________________________________________________________________________________________________

	Title: Agglomerative hierarchical clustering using the Receptor Noise Limited model
	''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
	Authors: Jolyon Troscianko & Cedric van den Berg
	Date: 28/7/2015
............................................................................................................

This code uses agglomerative hierarchical clustering to reduce a colour image down to a 'usable'
number of clusters. The code starts with each pixel being its own cluster. Each cluster is compared
to its neighbouring clusters in the image XY plane, and distances calculated from V&O RNL distances
(JNDs).

Following distance measurements each cluster is combined with its nearest neighbouring cluster.
Nodes can have multiple clustering events at each pass, e.g. if cluster A is closest to B, but B is closest
to C then alll three will be clustered, meaning whole strings or neighbouring regions can be clustered.
In practice each cluster tends to be combined with two or three clusters on each pass.

The distance radius increases with each pass, so as clusters get larger they can also combine with
neighbours further away. This also keeps the processing for each pass fairly constant (i.e. there
are fewer clusters on each pass, but each one must be compared to a larger number of neighbours).

Clustering therefore takes place across n + 2 dimensions (n colours plus x & y space).

After a specified number of passes it becomes more compuationally efficient to compare all clusters
to all clusters. This tends to be around pass 5-6. 

--------------------------------------------------------------------------------------------------------------------------------
Requirements:

Input images should be cone-catch images. The Weber fraction order must match the slice order.

The first two to four channels must be chromatic cone classes (e.g. lw, mw, sw, uv), the final channel
must be the luminance channel (e.g. double cone for birds, or the average of mw & lw for humans)
_________________________________________________________________________________________________
*/



import ij.*;
import ij.plugin.filter.PlugInFilter;
import ij.process.*;
import ij.gui.*;

import ij.measure.ResultsTable;
import ij.text.TextWindow;
import ij.plugin.filter.Analyzer;
import ij.WindowManager;

public class RNL_Clustering implements PlugInFilter {

ImageStack stack;
	public int setup(String arg, ImagePlus imp) { 
	stack = imp.getStack(); 
	
	return DOES_32 + STACK_REQUIRED; 
	}



public void run(ImageProcessor ip) {


IJ.showStatus("Clustering");

int w = stack.getWidth();
int h = stack.getHeight();
int nPoints = w*h;
int nDims = stack.getSize();

int nLoops = 20;
double radiusMultiplier = 2.0;
int haltClusters = 1;
int recordClusters = 20;
String imTitle = "ImageID";

double[] weber = new double[nDims]; //

if(nDims == 4){
	weber[0] = 0.05; // LW HUMAN scaled to 0.05 for LWS from Hofer et al 2005, assuming a 2:1 L:M and 5.72% S cones
	weber[1] = 0.0707106781; // MW
	weber[2] = 0.1657433633; // SW
	weber[3] = 0.1; // 
} else {
for(int i = 0; i<nDims; i++)
	weber[i] = 0.05;

}

double lumWeber = 0.05;

double JNDthreshold = 3.0;
double lumJNDthreshold = 3.0;
int minClusterSize = 10;
int nSwitch = 6;

String[] sliceNames = new String[nDims];
sliceNames = stack.getSliceLabels();

	GenericDialog gd = new GenericDialog("Weber Clustering");
		gd.addNumericField("Colour_JND_Threshold", JNDthreshold, 3);
		gd.addNumericField("Luminance_JND_Threshold", lumJNDthreshold, 3);
		gd.addMessage("Number of clustering loops");
		gd.addNumericField("Loops", nLoops, 0);
		gd.addMessage("The search radius is increased with each loop by this variable");
		gd.addNumericField("Radius Multiplier:", radiusMultiplier, 3);
		gd.addNumericField("Minimum cluster size", minClusterSize, 0);
		gd.addNumericField("Compare all clusters after pass", nSwitch, 0);
		gd.addMessage("Clustering will stop early if this value >1");
		gd.addNumericField("Stop Number:", haltClusters, 0);
		gd.addNumericField("Record from pass:", recordClusters, 0);
		//gd.addCheckbox("Output Image", true);
		gd.addStringField("Image ID:", imTitle, 20);
		gd.addMessage(" ");
		gd.addMessage("Weber Fractions");
		//gd.addNumericField("Weber_" + (i+1), weber[i], 3);

		for(int i=0; i<nDims-1; i++)
			gd.addNumericField("Weber_" + (i+1) + " " + sliceNames[i], weber[i], 3);	
		

		gd.addNumericField("Luminance_Weber_Fraction", lumWeber, 3);
		//gd.addCheckbox("Horizontal and vertical adjacency results", false);
		//gd.addCheckbox("Output_Adjacency_Matrix", false);

	gd.showDialog();
	if (gd.wasCanceled())
		return;

JNDthreshold = gd.getNextNumber();
lumJNDthreshold = gd.getNextNumber();

// ----------New distance based on both values--------------
//JNDthreshold = pow((JNDthreshold*JNDthreshold) + (lumJNDthreshold+lumJNDthreshold), 0.5);

nLoops = (int) gd.getNextNumber();
radiusMultiplier = (double) gd.getNextNumber();
minClusterSize = (int) gd.getNextNumber();
nSwitch = (int) gd.getNextNumber();
haltClusters = (int) gd.getNextNumber();
recordClusters = (int) gd.getNextNumber();

if(recordClusters>nLoops)
	recordClusters = nLoops;

imTitle = gd.getNextString();
for(int i=0; i<nDims-1; i++)
	weber[i] = gd.getNextNumber();
lumWeber = gd.getNextNumber();
//boolean outputHVdata = gd.getNextBoolean();
//boolean outMatrix = gd.getNextBoolean();

//boolean outImage = gd.getNextBoolean();


	IJ.log("_____________________________________________");
	IJ.log("----------------Weber Clustering----------------");
	IJ.log("colJND: " + JNDthreshold );
	IJ.log("lumJND: " + lumJNDthreshold);
	IJ.log("nLoops: " + nLoops);
	IJ.log("radiusMultiplier: " + radiusMultiplier);
	IJ.log("minClusterSize: " + minClusterSize);
	IJ.log("Switch to nxn comparisons: " + nSwitch);
	IJ.log("halt on n clusters: " + haltClusters);
	IJ.log("Record from pass: " + recordClusters);
	IJ.log("Label: " + imTitle);
	for(int i=0; i<nDims-1; i++)
		IJ.log("weber "+ sliceNames[i] + ": " + weber[i] );
	IJ.log("lumWeber: " + lumWeber);
	//IJ.log("Output Hrz&Vrt: " + outputHVdata);


	IJ.log("_____________________________________________");


double[] dfis = new double[nDims];
double JND = 0.0;
double lumJND = 0.0;

ImageStack cJNDStack = new ImageStack(w, h);
ImageStack lJNDStack = new ImageStack(w, h);

ImageStack outStack = new ImageStack(w, h);
ImageStack idStack = new ImageStack(w, h);

double pcCover = 0.0;
//int row = 0;
//ResultsTable rt = new ResultsTable();
//ResultsTable rt =  ResultsTable.getResultsTable();

double[] means = new double[nPoints*nDims];
double[] sds = new double[nPoints*nDims]; // note this code uses variance, not SDs for slightly faster processing
int[] ns = new int[nPoints]; // sample size counts
int[] clusterIDs = new int[nPoints];
int[] clusterLookups = new int[nPoints];

float[] pixels= new float[w*h]; // temp array for loading

for(int i=0; i<nPoints; i++){
	clusterIDs[i] = i; // set up pixel as a cluster
	ns[i] = 1;
	clusterLookups[i] = -1;
}


for(int i=0; i<nDims; i++){ // convert float pixel values to double-precision array
	pixels = (float[]) stack.getPixels(i+1);
	for(int k=0; k<w*h; k++)
		if(pixels[k] > 0) // don't import negative numbers
			means[(i*w*h)+k] = (double) (pixels[k]);
		else {
			means[k] = 0; // flag negative numbers in the first channel
			clusterIDs[k] = -1;
		}
}

for(int i=0; i<nDims*nPoints; i++)
	sds[i] = 1.0; // set initial variance to 1

int countPoints = 0;
for(int i=0; i<nPoints; i++)
	if(clusterIDs[i] != -1)
		countPoints ++;

int nClusters = countPoints;


// Arrays to hold the bounds of each cluster to speed up searching
int[] cbT = new int[nPoints]; // cluster bounds top
int[] cbB = new int[nPoints]; // bottom
int[] cbL = new int[nPoints]; // left
int[] cbR = new int[nPoints]; // right
double[] dists = new double[nPoints];
double[] lumDists = new double[nPoints];
int[] nearestID = new int[nPoints];

for(int y=0; y<h; y++)
	for(int x=0; x<w; x++){
		cbT[(y*w)+x] = y;
		cbB[(y*w)+x] = y;
		cbL[(y*w)+x] = x;
		cbR[(y*w)+x] = x;
	}//x

// Set up variables
double radius = 1.0;
int j = 0;
int prevID = -1;
double distSum = 0.0;
int iLoc = 0;
int jLoc = 0;
double p0 = 0.0;
double p1 = 0.0;
int oldLoc = 0;
int newLoc = 0;
int N = 0;
double GM = 0.0;
double ESS = 0.0;
double TGSS = 0.0;
int rT = 0;
int tMCS = 0;

//-----------------------------Initialise Results Table----------------------------

String[] titles = WindowManager.getNonImageTitles();
int cResultsOpen = 0;
int sResultsOpen = 0;

for(int i=0; i<titles.length; i++){
	if(titles[i] == "Summary Results")
		sResultsOpen ++;
	if(titles[i] == "Cluster Results")
		cResultsOpen ++;
}

if(cResultsOpen > 0)
	IJ.renameResults("Cluster Results", "Results");

ResultsTable ct = Analyzer.getResultsTable();
if (ct == null) {
	ct = new ResultsTable();
	Analyzer.setResultsTable(ct);
}


// ----------------------FIND EACH CLUSTER'S NEAREST NEIGHBOUR---------------------------

for(int m=0; m<nLoops; m++){

IJ.showProgress((float) m/nLoops) ;
//IJ.showStatus("Pass: " + (m+1) + ", clusters: " + (nClusters - (nPoints-countPoints)) );

for(int i=0; i<nPoints; i++){
	dists[i] = 10E10;
	lumDists[i] = 10E10;
	nearestID[i] = i;
}

rT = (int) (Math.round(radius));

if(m < nSwitch){

for(int i=0; i<nPoints; i++)
if(clusterIDs[i] == i){
	
	for(int y=cbT[i]-rT; y<=cbB[i]+rT; y++) // window bounds
	for(int x=cbL[i]-rT; x<=cbR[i]+rT; x++)
	if(x>=0 && x<w && y>=0 && y<h){ // make sure it's not off the image
		j = (y*w)+x;
		prevID = -1;	

		if( clusterIDs[j] != i && clusterIDs[j] != prevID) // don't compare to itself, and don't compare if it's the same cluster as the previous pass
		//if( clusterIDs[j] != i)
		if( nearestID[i] != j){ // don't compare if it's already the nearest neighbour
			
			prevID = clusterIDs[j];

			for(int k=0; k<nDims-1; k++)
				dfis[k] = Math.log(means[(k*nPoints)+i]/means[(k*nPoints)+clusterIDs[j]]);

			if(nDims == 3) // DICHROMATIC EQUATION
				JND = Math.pow(Math.pow(dfis[0]-dfis[1],2)/(Math.pow(weber[0],2)+Math.pow(weber[1],2)),0.5);

			else if(nDims == 4) // TRICHROMATIC EQUATION
				JND = Math.pow((Math.pow(weber[0],2)*Math.pow(dfis[2]-dfis[1],2)+Math.pow(weber[1],2)*Math.pow(dfis[2]-dfis[0],2)+Math.pow(weber[2],2)*Math.pow(dfis[0]-dfis[1],2))/(Math.pow(weber[0]*weber[1],2)+Math.pow(weber[0]*weber[2],2)+Math.pow(weber[1]*weber[2],2)),0.5);

			else if(nDims == 5) // TETRACHROMATIC EQUATION
				JND = Math.pow((Math.pow(weber[0]*weber[1],2)*Math.pow(dfis[3]-dfis[2],2)+Math.pow(weber[0]*weber[2],2)*Math.pow(dfis[3]-dfis[1],2)+Math.pow(weber[0]*weber[3],2)*Math.pow(dfis[2]-dfis[1],2)+Math.pow(weber[1]*weber[2],2)*Math.pow(dfis[3]-dfis[0],2)+Math.pow(weber[1]*weber[3],2)*Math.pow(dfis[2]-dfis[0],2)+Math.pow(weber[2]*weber[3],2)*Math.pow(dfis[1]-dfis[0],2))/(Math.pow(weber[0]*weber[1]*weber[2],2)+Math.pow(weber[0]*weber[1]*weber[3],2)+Math.pow(weber[0]*weber[2]*weber[3],2)+Math.pow(weber[1]*weber[2]*weber[3],2)),0.5);

			else {
				IJ.log("Image must have 2, 3 or 4 channels plus luminance in the last channel");
				return;
			}


			// --------------------LUMINANCE JND -------------
			// Siddiqi JNDs:
			lumJND=Math.abs((Math.log(means[((nDims-1)*nPoints)+i])-Math.log(means[((nDims-1)*nPoints)+clusterIDs[j]]))/lumWeber);
			// Endler JNDs:
			//lumJND = Math.abs((Math.log(means[((nDims-1)*nPoints)+i])-Math.log(means[((nDims-1)*nPoints)+clusterIDs[j]]))/(Math.log(means[((nDims-1)*nPoints)+i])+Math.log(means[((nDims-1)*nPoints)+clusterIDs[j]])))/lumWeber; 

			//---------------Combine luminance and colour JNDs into a single distance weighted by JND thresholds--------------
			//JND = pow((JND*JND) + (lumJND*lumJND),0.5);

			JND = Math.pow( Math.pow(JND/JNDthreshold, 2) + Math.pow(lumJND/lumJNDthreshold, 2) , 0.5);

			if(JND !=0)
			if(JND< dists[i]){
				dists[i] = JND;
				lumDists[i] = lumJND;
				nearestID[i] = j;
			}//fi
		}// fi
	} //

}


} else { // if number of clusters is below threshold, compare all clusters to all clusters.



for(int i=0; i<nPoints; i++)
if(clusterIDs[i] == i)
for(j = 0; j<nPoints; j++) 
if(j!=i && clusterIDs[j] == j)
if( nearestID[i] != j){ // don't compare if it's already the nearest neighbour

			for(int k=0; k<nDims-1; k++)
				dfis[k] = Math.log(means[(k*nPoints)+i]/means[(k*nPoints)+clusterIDs[j]]);

			if(nDims == 3) // DICHROMATIC EQUATION
				JND = Math.pow(Math.pow(dfis[0]-dfis[1],2)/(Math.pow(weber[0],2)+Math.pow(weber[1],2)),0.5);

			else if(nDims == 4) // TRICHROMATIC EQUATION
				JND = Math.pow((Math.pow(weber[0],2)*Math.pow(dfis[2]-dfis[1],2)+Math.pow(weber[1],2)*Math.pow(dfis[2]-dfis[0],2)+Math.pow(weber[2],2)*Math.pow(dfis[0]-dfis[1],2))/(Math.pow(weber[0]*weber[1],2)+Math.pow(weber[0]*weber[2],2)+Math.pow(weber[1]*weber[2],2)),0.5);

			else if(nDims == 5) // TETRACHROMATIC EQUATION
				JND = Math.pow((Math.pow(weber[0]*weber[1],2)*Math.pow(dfis[3]-dfis[2],2)+Math.pow(weber[0]*weber[2],2)*Math.pow(dfis[3]-dfis[1],2)+Math.pow(weber[0]*weber[3],2)*Math.pow(dfis[2]-dfis[1],2)+Math.pow(weber[1]*weber[2],2)*Math.pow(dfis[3]-dfis[0],2)+Math.pow(weber[1]*weber[3],2)*Math.pow(dfis[2]-dfis[0],2)+Math.pow(weber[2]*weber[3],2)*Math.pow(dfis[1]-dfis[0],2))/(Math.pow(weber[0]*weber[1]*weber[2],2)+Math.pow(weber[0]*weber[1]*weber[3],2)+Math.pow(weber[0]*weber[2]*weber[3],2)+Math.pow(weber[1]*weber[2]*weber[3],2)),0.5);

			else {
				IJ.log("Image must have 2, 3 or 4 channels plus luminance in the last channel");
				return;
			}


			// --------------------LUMINANCE JND -------------
			// Siddiqi JNDs:
			lumJND=Math.abs((Math.log(means[((nDims-1)*nPoints)+i])-Math.log(means[((nDims-1)*nPoints)+clusterIDs[j]]))/lumWeber);
			// Endler JNDs:
			//lumJND = Math.abs((Math.log(means[((nDims-1)*nPoints)+i])-Math.log(means[((nDims-1)*nPoints)+clusterIDs[j]]))/(Math.log(means[((nDims-1)*nPoints)+i])+Math.log(means[((nDims-1)*nPoints)+clusterIDs[j]])))/lumWeber; 

			//---------------Combine luminance and colour JNDs into a single distance weighted by JND thresholds--------------
			//JND = pow((JND*JND) + (lumJND*lumJND),0.5);

			JND = Math.pow( Math.pow(JND/JNDthreshold, 2) + Math.pow(lumJND/lumJNDthreshold, 2) , 0.5);

			if(JND !=0)
			if(JND< dists[i]){
				dists[i] = JND;
				lumDists[i] = lumJND;
				nearestID[i] = j;
			}//fi

	}



}// else threshold number of clusters

//--------------------------COMBINE NEAREST CLUSTERS--------------------------------


tMCS = (int) (Math.round(minClusterSize * (1.0*(m+1)/nLoops))); // minimum clusters size ramps up to the full number by the final pass
//IJ.showStatus("Pass: " + (m+1) + ", clusters: " + (nClusters - (nPoints-countPoints)) + ", radius: " + Math.round(radius) + ", min size:" + tMCS);
IJ.showStatus("Pass: " + (m+1) + ", clusters: " + nClusters + ", radius: " + Math.round(radius) + ", min size:" + tMCS);

for(int i=0; i<nPoints; i++)
if(clusterIDs[i] == i)  // select only cluster reference positions
if(dists[i] < 1.4142135624 || ns[i]<tMCS)
if(i != clusterIDs[nearestID[i]]){ // ignore if the nearest cluster is itself

		nClusters --;
		oldLoc = clusterIDs[nearestID[i]]; // use lower value as cluster ID
		newLoc = i;

		N = ns[newLoc] + ns[oldLoc];

		for(int k=0; k<nDims; k++){ // combine means & variances (note variance is used, not SD)
			GM = ( (means[(k*nPoints)+newLoc] * ns[newLoc]) + (means[(k*nPoints)+oldLoc]) * ns[oldLoc]) / N;
			ESS = (sds[(k*nPoints)+newLoc] * (ns[newLoc] - 1)) + (sds[(k*nPoints)+oldLoc] * (ns[oldLoc] - 1));
			TGSS = (Math.pow(means[(k*nPoints)+newLoc] - GM, 2) * ns[newLoc]) + (Math.pow(means[(k*nPoints)+oldLoc] - GM, 2) * ns[oldLoc]);
			sds[(k*nPoints)+newLoc] = (ESS+TGSS) / (N-1);
			means[(k*nPoints)+newLoc] = GM;
		}//k
		ns[newLoc] = N;

		for(int y=cbT[oldLoc]; y<=cbB[oldLoc]; y++) // update cluster IDs
		for(int x=cbL[oldLoc]; x<=cbR[oldLoc]; x++)
			if(clusterIDs[(y*w)+x] == oldLoc)
				clusterIDs[(y*w)+x] = newLoc;


	// Update bounds of enlarged cluster
		if(cbT[oldLoc] < cbT[newLoc]) // TOP
			cbT[newLoc] = cbT[oldLoc];

		if(cbB[oldLoc] > cbB[newLoc]) // BOTTOM
			cbB[newLoc] = cbB[oldLoc];

		if(cbR[oldLoc] > cbR[newLoc]) // RIGHT
			cbR[newLoc] = cbR[oldLoc];

		if(cbL[oldLoc] < cbL[newLoc]) // LEFT
			cbL[newLoc] = cbL[oldLoc];


	// Update the nearest-neighbour of the oldLoc
	// if the oldLoc was due to be linked to a 3rd group then re-loop this iteration, adding 3rd loc (this can then start a whole chain)

	if(nearestID[oldLoc] != newLoc){ // oldLoc's nearest neighbour was a 3rd value, so add this as newLoc's nearest neighbour & re-loop
		nearestID[i] = nearestID[oldLoc];
		i = i-1;
	}
} // i 

if(m>recordClusters-2 || nClusters == 1){


	int[] clusterCodes = new int[nClusters];
	for(j=0; j<nClusters; j++)
		clusterCodes[j] = -1; // fill array with -1s

	//clusterCodes[0] = clusterIDs[0];

	//int clusterCount = 0;
	int clusterCount = -1;
	int clusterMatch = 0;

	for(int i=0; i<nPoints; i++)
	if(clusterIDs[i] != -1){
		clusterMatch = 0;
		for(j=0; j<nClusters; j++){
			if(clusterIDs[i] == clusterCodes[j]){
				clusterLookups[i] = j;
				clusterMatch ++;
			}
		}//j

		if(clusterMatch == 0){ // novel cluster ID found
			clusterCount++;
			//if(clusterCount>=nClusters) // this isn't needed
			//	clusterCount = nClusters-1;
			clusterCodes[clusterCount] = clusterIDs[i];
			clusterLookups[i] = clusterCount;
		}
	}//i


	//--------------Chomraticity variables--------------

	double Dtemp = 0.0;
	String DmaxChannel = "";
	double[] Dmaxs = new double[nClusters];
	//double[] Daves = new double[nClusters];
	double[] Lums = new double[nClusters];
	double[] prCovers = new double[nClusters];

	double[] cMeans = new double[nClusters * (nDims-1)];

	double rnlX = 0.0;
	double rnlY = 0.0;
	double rnlZ = 0.0;
	double A = 0.0;
	double a = 0.0;
	double b = 0.0;
	double c = 0.0;
	double[] RNLChr = new double[nClusters];
	double c0 = 0.0;
	double c1 = 0.0;
	double c2 = 0.0;
	double c3 = 0.0;



	//---------------------------------RECORD RESULTS--------------------------------
	//row = 0;
	for(int k=0; k<nClusters; k++)
	for(int i=0; i<nPoints; i++)
		if(clusterIDs[i] == i && clusterLookups[i] == k){

			ct.incrementCounter();
			ct.addValue("Image", imTitle);
			ct.addValue("Pass", (m+1));
			ct.addValue("ClusterID", clusterLookups[i]+1);	

			for(j=0; j<nDims; j++){
				ct.addValue(sliceNames[j] + "_mean", means[(nPoints*j)+i]);
				ct.addValue(sliceNames[j] + "_sd", Math.pow(sds[(nPoints*j)+i],0.5));
			}

			for(j=0; j<nDims-1; j++)
				cMeans[(k*(nDims-1))+j] = means[(nPoints*j)+i];

			ct.addValue("Area", ns[i]);
			prCovers[k] = (double) ((double) (ns[i])/(double) (countPoints));
			ct.addValue("Coverage (pc)", (float) (prCovers[k]*100.0));

			//--------------Calculate RNL Chromaticity----------------------

			RNLChr[k] = 0.0;

			if(nDims >= 3){ // DICHROMAT
				c0 = means[i] ;
				c1 = means[nPoints+i] ;
				rnlX =  Math.pow(1/(Math.pow(weber[1],2)+Math.pow(weber[0], 2)), 0.5)*(Math.log(c0)-Math.log(c1));	
				RNLChr[k] += (rnlX * rnlX);
				ct.addValue("RNL X", rnlX);
			}
			if(nDims >= 4){ // TRICHROMAT
				c2 = means[(2*nPoints)+i] ;
				rnlY = Math.pow((Math.pow(weber[1] , 2)+Math.pow(weber[0] , 2))/(Math.pow(weber[2]*weber[1], 2)+Math.pow(weber[2]*weber[0], 2)+Math.pow(weber[1]*weber[0],2)) , 0.5)*(Math.log(c2)-(Math.log(c0)*(Math.pow(weber[1], 2)/(Math.pow(weber[1], 2)+Math.pow(weber[0],2)))+Math.log(c1)*(Math.pow(weber[0],2)/(Math.pow(weber[1], 2)+Math.pow(weber[0], 2)))))      ;
				RNLChr[k] += (rnlY * rnlY);
				ct.addValue("RNL Y", rnlY);
			}
			if(nDims >= 5){ // TETRACHROMAT
				c3 = means[(3*nPoints)+i] ; 
				A = Math.pow((Math.pow(weber[1]*weber[0], 2)+Math.pow(weber[2]*weber[0], 2)+Math.pow(weber[2]*weber[1], 2))/(Math.pow(weber[2]*weber[1]*weber[0], 2)+Math.pow(weber[3]*weber[1]*weber[0], 2)+Math.pow(weber[3]*weber[2]*weber[0], 2)+Math.pow(weber[3]*weber[2]*weber[1], 2)),0.5);
				a = Math.pow(weber[2]*weber[1], 2)/(Math.pow(weber[1]*weber[0], 2)+Math.pow(weber[2]*weber[1], 2)+Math.pow(weber[2]*weber[0], 2));
				b = Math.pow(weber[2]*weber[0], 2)/(Math.pow(weber[1]*weber[0], 2)+Math.pow(weber[2]*weber[1], 2)+Math.pow(weber[2]*weber[0], 2));
				c = Math.pow(weber[1]*weber[0], 2)/(Math.pow(weber[1]*weber[0], 2)+Math.pow(weber[2]*weber[1], 2)+Math.pow(weber[2]*weber[0], 2));
				rnlZ =      A*(Math.log(c3)-(a*Math.log(c0)+b*Math.log(c1)+c*Math.log(c2)))     ;
				RNLChr[k] += (rnlZ * rnlZ);
				ct.addValue("RNL Z", rnlZ);
			}

			RNLChr[k] = Math.pow(RNLChr[k], 0.5);
			ct.addValue("RNL Saturation", RNLChr[k]);
			

			//---------------Calculate Chromaticity-----------------
			

			//Dmax = 0.0;
			Dtemp = 0.0;
			//Dave = 0.0;
			Dmaxs[k] = 0; // these oponent chanels are on a -1 to 1 range, where zero is grey
			for(j=0; j<nDims-1; j++)	
			for(int jj=j+1; jj<nDims-1; jj++){
				Dtemp = Math.abs(   (means[(nPoints*j)+i] - means[(nPoints*jj)+i]) / (means[(nPoints*j)+i] + means[(nPoints*jj)+i])  );
				if(Dtemp > Dmaxs[k]){
					Dmaxs[k] = Dtemp;
					DmaxChannel = sliceNames[j] +":"+sliceNames[jj];
				}
				//RNLChr[k] += Dtemp;
			}
			//RNLChr[k] = RNLChr[k]/(nDims-1); // average distance from grey

			ct.addValue("Dmax Chromaticity", Dmaxs[k]);
			ct.addValue("Dmax channel", DmaxChannel);
		//	ct.addValue("Dave Chromaticity", RNLChr[k]);

			Lums[k] = means[(nPoints*(nDims-1))+i];
			
		} // i&k

	ct.showRowNumbers(false);
	ct.show("Results");

	IJ.renameResults("Results", "Cluster Results");

	// OUTPUT MEANS IN NEW 32-BIT IMAGE STACK

	for(int i=0; i<nDims; i++){
		float[] fPixels= new float[w*h]; // re-initialise array
		for(int k=0; k<w*h; k++)
			if(clusterIDs[k] != -1)
				fPixels[k] = (float) (means[(i*w*h)+clusterIDs[k]]);

		outStack.addSlice("Pass" + (m+1) + "_" + sliceNames[i], fPixels);
	}//i

	// ALSO OUTPUT CLUSTER IDs
	float[] fPixels= new float[w*h]; // re-initialise array
	for(int k=0; k<w*h; k++)
		//if(means[k] != 0)
		if(clusterIDs[k] != -1)
			//fPixels[k] = (float) (clusterIDs[k]);
			fPixels[k] = (float) (clusterLookups[k]+1);
	idStack.addSlice("Pass" + (m+1) + "_Cluster_IDs", fPixels);

}// m record clusters


if(nClusters <= haltClusters){
	m=nLoops-1;
	IJ.log("Stopping - reached cluster limit");
}


if(pcCover == 1.0)
	m=nLoops;

if(radius < w || radius < h) // expand the radius unless it's bigger than the image itself
	radius = radiusMultiplier * radius;


}//m


new ImagePlus(imTitle + "_Clustered", outStack).show();
new ImagePlus(imTitle + "_Cluster_IDs", idStack).show();

}
}
